/* 
 * Copyright (c) 2009, Artem Shinkarov <artyom.shinkaroff@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * This is library for Array of bits.
 * Arrays are implemented using uint8_t arrays.
 * Library alow to have a random access to the array,
 * reading and writing N bits in the location specified.
 *
 * Also the bit array can be considered as stream for
 * reading or writing.
 */
#ifndef __BITS_H__
#define __BITS_H__

#include <sys/types.h>

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>

/* Bits library Error Codes */
#define BITS_OK             0
#define BITS_CANNOTREAD     1
#define BITS_SIZEERROR      2

extern const char *bits_error_table[3];

#ifdef __DEBUG__
#define dprintf(fmt, ...)                               \
    (void)fprintf(stderr, "%s, %4u:\t" fmt,             \
                  __FILE__, __LINE__, ##__VA_ARGS__)
#else
#define dprintf(fmt, ...)   /* Nothing */
#endif


#define GET_BITS(__res, __bitptr, __num, __error)       \
    do {                                                \
        __res = get_bits(__bitptr, __num, __error);     \
        /*if (*__error != BITS_OK)                        \
            err(-1, "BITS Library ERROR:%s:%4u, %s\n",  \
                __FILE__, __LINE__,                     \
                bits_error_table[*__error]); */           \
    } while (0)

/*
 * Structure describing bit array.
 */
struct bits {
    uint8_t     *buf;     // buffer with data
    uint8_t     *bufPtr;  // current character in buffer
    uint8_t      bitPtr;  // pointer how many bits from this
                          // byte we already read, values -- [0..7]

    size_t       size;    // size of data in a buffer
};


/* 
 * Function prototypes 
 */

/*@null@*/
struct bits     *init_bits  ();
void             free_bits  (struct bits *);
inline uint32_t __attribute__ ((always_inline)) _get_bits   (struct bits *, uint8_t, int *, uint8_t **, uint8_t *);
inline uint32_t  get_bits   (struct bits *, uint8_t, int *);
void            _write_ubits(struct bits *, uint8_t, uint8_t, int *, 
                             uint8_t **, uint8_t *);
void             write_ubyte(struct bits *, uint8_t, int *);
void             move_pointer(struct bits *, int, int *);
#endif // __BITS_H__
