/***************************************************************************
* aabb.cpp                                                                 *
*                                                                          *
* The AABB structure encodes a three-dimensional axis-aligned box as a     *
* collection of three intervals, one for each axis.  These boxes are ideal *
* as bounding boxes as they are equipped with a very efficient ray-box     *
* intersector.  There are also a number of useful methods & functions that *
* expand the box, transform it, compute its surface area, etc.             *
*                                                                          *
* History:                                                                 *
*   12/11/2004  Initial coding.                                            *
*                                                                          *
***************************************************************************/

#include"aabb.h"

double SurfaceArea( struct AABB A )
{
    if( aabbIsNull( A ) ) return 0.0;
    return 2.0 * ( Len(A.X) * ( Len(A.Y) + Len(A.Z) ) + Len(A.Y) * Len(A.Z) );
}

bool aabbIsNull( struct AABB A )
{ 
    return intervalIsNull( A.X ) || intervalIsNull( A.Y ) || intervalIsNull( A.Z );
}

bool aabbExpand( struct AABB *a, struct AABB b )
{
    // Expand this box to enclose box "b" if it does not already do so.
    // Return true iff the box actually had to be expanded.
    bool expand = false;
    if( b.X.min < a->X.min ) { expand = true; a->X.min = b.X.min; } 
    if( b.X.max > a->X.max ) { expand = true; a->X.max = b.X.max; } 
    if( b.Y.min < a->Y.min ) { expand = true; a->Y.min = b.Y.min; } 
    if( b.Y.max > a->Y.max ) { expand = true; a->Y.max = b.Y.max; } 
    if( b.Z.min < a->Z.min ) { expand = true; a->Z.min = b.Z.min; } 
    if( b.Z.max > a->Z.max ) { expand = true; a->Z.max = b.Z.max; } 
    return expand;
}
