#include <math.h>
#include "vec3.h"

void vec3Set( struct Vec3 *v, double x, double y, double z )
{
	v->x = x;
	v->y = y;
	v->z = z;
}

double LengthSquared( struct Vec3 A )
{
    return A.x * A.x + A.y * A.y + A.z * A.z;
}

struct Vec3 vec3Minus( struct Vec3 A )
{
	struct Vec3 v;
	v.x = -A.x;
	v.y = -A.y;
	v.z = -A.z;
	return v;
}

struct Vec3 vec3Add( struct Vec3 A, struct Vec3 B )
{
	struct Vec3 v;
	v.x = A.x + B.x;
	v.y = A.y + B.y;
	v.z = A.z + B.z;
    return v;
}

struct Vec3 vec3Sub( struct Vec3 A, struct Vec3 B )
{
	struct Vec3 v;
	v.x = A.x - B.x;
	v.y = A.y - B.y;
	v.z = A.z - B.z;
    return v;
}

struct Vec3 vec3Mul( double a, struct Vec3 A )
{
	struct Vec3 v;
	v.x = a * A.x;
	v.y = a * A.y;
	v.z = a * A.z;
	return v;
}

struct Vec3 vec3Div( struct Vec3 A, double c )
{
	struct Vec3 v;
	v.x = A.x / c;
	v.y = A.y / c;
	v.z = A.z / c;
	return v;
}

struct Vec3 Unit( struct Vec3 A )
{
    double d = LengthSquared( A );
	struct Vec3 v0;
	v0.x = 0; v0.y = 0; v0.z = 0;

	if(d == 0) 
		return v0;
    else
		return vec3Div( A, sqrt(d) );
}

double vec3InnerProduct( struct Vec3 A, struct Vec3 B )  // Inner product.
{
    return (A.x * B.x) + (A.y * B.y) + (A.z * B.z);
}

struct Vec3 vec3CrossProduct( struct Vec3 A, struct Vec3 B ) // Cross product.
{
	struct Vec3 v;
	v.x = A.y * B.z - A.z * B.y;
	v.y = A.z * B.x - A.x * B.z;
	v.z = A.x * B.y - A.y * B.x;
    return v;
}

struct Vec3 vec3RCP( struct Vec3 A, struct Vec3 B )  // Remove component parallel to B.
{
    double x = LengthSquared( B );
    if( x > 0.0 ) return vec3Sub(A, vec3Mul( vec3InnerProduct( A, B ) / x, B ) );
	return A;
}

double Length( struct Vec3 A )
{
    return sqrt( LengthSquared( A ) );
}

double dist( struct Vec3 A, struct Vec3 B ) // Euclidean distance from A to B.
{ 
    return Length( vec3Sub( A, B ) ); 
}
